@@ -35,7 +35,7 @@
 #include "file.h"
 #include "cgi.h"
 
-static void conn_send(struct uh_connection *conn, const void *data, ssize_t len)
+static void conn_send(struct uh_connection *conn, const void *data, size_t len)
 {
     struct uh_connection_internal *conni = (struct uh_connection_internal *)conn;
 
@@ -46,7 +46,7 @@ static void conn_send(struct uh_connection *conn, const void *data, ssize_t len)
     ev_io_start(conni->l->srv->loop, &conni->iow);
 }
 
-static void conn_send_chunk(struct uh_connection *conn, const void *data, ssize_t len)
+static void conn_send_chunk(struct uh_connection *conn, const void *data, size_t len)
 {
     struct uh_connection_internal *conni = (struct uh_connection_internal *)conn;
     struct buffer *wb = &conni->wb;
@@ -54,7 +54,7 @@ static void conn_send_chunk(struct uh_connection *conn, const void *data, ssize_
     if (unlikely(conni->flags & CONN_F_CLOSED))
         return;
 
-    buffer_put_printf(wb, "%X\r\n", (unsigned int)len);
+    buffer_put_printf(wb, "%zX\r\n", len);
     buffer_put_data(wb, data, len);
     buffer_put_data(wb, "\r\n", 2);
     ev_io_start(conni->l->srv->loop, &conni->iow);
@@ -121,6 +121,7 @@ static void conn_send_header_v(struct uh_connection *conn, const char *name, con
     buffer_put_printf(wb, "%s: ", name);
     buffer_put_vprintf(wb, value, arg);
     buffer_put_data(wb, "\r\n", 2);
+    ev_io_start(conni->l->srv->loop, &conni->iow);
 }
 
 static inline void conn_send_header(struct uh_connection *conn, const char *name, const char *value, ...)
@@ -137,6 +138,7 @@ static inline void conn_end_headers(struct uh_connection *conn)
     struct uh_connection_internal *conni = (struct uh_connection_internal *)conn;
 
     buffer_put_data(&conni->wb, "\r\n", 2);
+    ev_io_start(conni->l->srv->loop, &conni->iow);
 
     if (conni->resp.chunked) {
         conn->send = conn_send_chunk;
@@ -152,11 +154,17 @@ static inline void conn_end_headers(struct uh_connection *conn)
 static void conn_send_head_v(struct uh_connection *conn, int code, int64_t content_length, const char *reason, va_list arg)
 {
     struct uh_connection_internal *conni = (struct uh_connection_internal *)conn;
+    const struct uh_str path = conn->get_path(conn);
     struct buffer *wb = &conni->wb;
+    char addr_str[INET6_ADDRSTRLEN];
+    int port;
 
     if (likely(!reason))
         reason = http_status_str(code);
 
+    log_info("%s %d  %s %.*s - %d %s\n", addr_str, (saddr2str(&conni->paddr.sa, addr_str, sizeof(addr_str), &port) ? port : 0),
+        http_method_str(conni->parser.method), (int)path.len, path.p, code, reason);
+
     buffer_put_printf(wb, "HTTP/1.1 %d ", code);
     buffer_put_vprintf(wb, reason, arg);
     buffer_put_printf(wb, "\r\nServer: Libuhttpd/%s\r\n", UHTTPD_VERSION_STRING);
@@ -173,6 +181,8 @@ static void conn_send_head_v(struct uh_connection *conn, int code, int64_t conte
 
     conn->send_header = conn_send_header;
     conn->end_headers = conn_end_headers;
+
+    ev_io_start(conni->l->srv->loop, &conni->iow);
 }
 
 static inline void conn_send_head(struct uh_connection *conn, int code, int64_t content_length, const char *reason, ...)
@@ -242,8 +252,6 @@ static void conn_send_redirect(struct uh_connection *conn, int code, const char
 {
     va_list arg;
 
-    assert((code == HTTP_STATUS_MOVED_PERMANENTLY || code == HTTP_STATUS_FOUND) && location);
-
     conn_send_head(conn, code, 0, NULL);
 
     va_start(arg, location);
@@ -254,6 +262,86 @@ static void conn_send_redirect(struct uh_connection *conn, int code, const char
     conn_end_response(conn);
 }
 
+static bool conn_https_redirect(struct uh_connection *conn)
+{
+    struct uh_connection_internal *conni = (struct uh_connection_internal *)conn;
+    struct uh_str host = conn->get_header(conn, "host");
+    struct uh_str url = conn->get_uri(conn);
+    struct sockaddr_in6 sin6;
+    struct uh_listener *l;
+    int tls_port = -1;
+    int host_len = 0;
+    const char *p;
+
+    if (conni->l->ssl)
+        return false;
+
+    list_for_each_entry(l, &conni->l->srv->listeners, list) {
+        socklen_t sl = sizeof(struct sockaddr_in6);
+
+        if (!l->ssl)
+            continue;
+
+        getsockname(l->sock, (struct sockaddr *)&sin6, &sl);
+
+        if (sin6.sin6_family != conni->saddr.sa.sa_family)
+            continue;
+
+        if (tls_port != -1 && ntohs(sin6.sin6_port) != 443)
+			continue;
+
+		tls_port = ntohs(sin6.sin6_port);
+    }
+
+    if (tls_port == -1)
+        return false;
+
+    if (host.len == 0)
+        return false;
+
+    host_len = 0;
+    p = host.p;
+
+    while (p < host.p + host.len) {
+        if (*p++ == ']') {
+            host_len = p - host.p - 1;
+            break;
+        }
+    }
+
+    if (host_len == 0) {
+        p = host.p;
+        while (p < host.p + host.len) {
+            if (*p++ == ':') {
+                host_len = p - host.p - 1;
+                break;
+            }
+        }
+    }
+
+    if (host_len == 0)
+        host_len = host.len;
+
+    if (tls_port == 443)
+        conn->send_redirect(conn, HTTP_STATUS_TEMPORARY_REDIRECT,
+            "https://%.*s%.*s", host_len, host.p, (int)url.len, url.p);
+    else
+        conn->send_redirect(conn, HTTP_STATUS_TEMPORARY_REDIRECT,
+            "https://%.*s:%d%.*s", host_len, host.p, tls_port, (int)url.len, url.p);
+
+    return true;
+}
+
+static void conn_check_expect_100_continue(struct uh_connection *conn)
+{
+    struct uh_str expect = conn->get_header(conn, "Expect");
+
+    if (uh_str_equal_case(&expect, "100-continue")) {
+        conn->send_head(conn, HTTP_STATUS_CONTINUE, 0, NULL);
+        conn->end_headers(conn);
+    }
+}
+
 static inline const struct sockaddr *conn_get_paddr(struct uh_connection *conn)
 {
     struct uh_connection_internal *conni = (struct uh_connection_internal *)conn;
@@ -535,11 +623,8 @@ static int on_headers_complete(struct http_parser *parser)
     struct uh_server_internal *srv = conn->l->srv;
     struct http_parser_url *u = &conn->url_parser;
     struct uh_request *req = &conn->req;
-    struct sockaddr *sa = &conn->paddr.sa;
     uh_path_handler_prototype handler;
-    char addr_str[INET6_ADDRSTRLEN];
     struct uh_str path;
-    int port;
 
     canonpath((char *)O2D(conn, req->url.offset), &req->url.length);
 
@@ -548,9 +633,6 @@ static int on_headers_complete(struct http_parser *parser)
     path.p = O2D(conn, u->field_data[UF_PATH].off) + req->url.offset;
     path.len = u->field_data[UF_PATH].len;
 
-    log_info("%s %d  %s %.*s\n", addr_str, (saddr2str(sa, addr_str, sizeof(addr_str), &port) ? port : 0),
-        http_method_str(parser->method), (int)path.len, path.p);
-
     handler = find_path_handler(conn, &srv->handlers, &path);
     if (!handler)
         handler = find_plugin_handler(conn, &srv->plugins, &path);
@@ -603,7 +685,8 @@ static int on_message_complete_cb(struct http_parser *parser)
 
     ev_timer_stop(srv->loop, &conn->timer);
 
-    conn->handler(&conn->com, UH_EV_COMPLETE);
+    if (!conn->l->srv->https_redirect || !conn_https_redirect(&conn->com))
+        conn->handler(&conn->com, UH_EV_COMPLETE);
 
     http_parser_pause(parser, true);
 
@@ -620,6 +703,13 @@ static struct http_parser_settings settings = {
     .on_message_complete = on_message_complete_cb
 };
 
+static bool conn_closed(struct uh_connection *conn)
+{
+    struct uh_connection_internal *conni = (struct uh_connection_internal *)conn;
+
+    return __sync_fetch_and_or(&conni->closed, 0);
+}
+
 static void conn_incref(struct uh_connection *conn)
 {
     struct uh_connection_internal *conni = (struct uh_connection_internal *)conn;
@@ -653,6 +743,11 @@ void conn_free(struct uh_connection_internal *conn)
 
     conn->flags |= CONN_F_CLOSED;
 
+    if (__sync_fetch_and_or(&conn->closed, 1)) {
+        conn_decref((struct uh_connection *)conn);
+        return;
+    }
+
     ev_timer_stop(loop, &conn->timer);
     ev_io_stop(loop, &conn->ior);
     ev_io_stop(loop, &conn->iow);
@@ -677,7 +772,7 @@ void conn_free(struct uh_connection_internal *conn)
 
     cgi_free(conn);
 
-    log_info("Connection(%s %d) closed\n", addr_str,
+    log_debug("Connection(%s %d) closed\n", addr_str,
             (saddr2str(&conn->paddr.sa, addr_str, sizeof(addr_str), &port) ? port : 0));
 
     conn_decref((struct uh_connection *)conn);
@@ -736,12 +831,12 @@ static int ssl_negotiated(struct uh_connection_internal *conn)
     char err_buf[128];
     int ret;
 
-    ret = ssl_connect(conn->ssl, true, on_ssl_verify_error, NULL);
+    ret = ssl_accept(conn->ssl, on_ssl_verify_error, NULL);
     if (ret == SSL_PENDING)
         return 0;
 
     if (ret == SSL_ERROR) {
-        log_err("ssl connect error(%d): %s\n", ssl_err_code, ssl_strerror(ssl_err_code, err_buf, sizeof(err_buf)));
+        log_err("ssl connect error: %s\n", ssl_last_error_string(err_buf, sizeof(err_buf)));
         return -1;
     }
 
@@ -758,8 +853,7 @@ static int conn_ssl_read(int fd, void *buf, size_t count, void *arg)
 
     ret = ssl_read(conn->ssl, buf, count);
     if (ret == SSL_ERROR) {
-        log_err("ssl_read(%d): %s\n", ssl_err_code,
-                ssl_strerror(ssl_err_code, err_buf, sizeof(err_buf)));
+        log_err("ssl_read: %s\n", ssl_last_error_string(err_buf, sizeof(err_buf)));
         return P_FD_ERR;
     }
 
@@ -790,8 +884,7 @@ static void conn_write_cb(struct ev_loop *loop, struct ev_io *w, int revents)
 
         ret = ssl_write(conn->ssl, buffer_data(b), buffer_length(b));
         if (ret == SSL_ERROR) {
-            log_err("ssl_write(%d): %s\n", ssl_err_code,
-                    ssl_strerror(ssl_err_code, err_buf, sizeof(err_buf)));
+            log_err("ssl_write: %s\n", ssl_last_error_string(err_buf, sizeof(err_buf)));
             goto err;
         }
 
@@ -970,6 +1063,9 @@ static void conn_init_cb(struct uh_connection *conn)
 
     conn->send_error = conn_send_error;
     conn->send_redirect = conn_send_redirect;
+    conn->https_redirect = conn_https_redirect;
+
+    conn->check_expect_100_continue = conn_check_expect_100_continue;
 
     conn->end_response = conn_end_response;
 
@@ -977,6 +1073,7 @@ static void conn_init_cb(struct uh_connection *conn)
     conn->download_file = download_file;
     conn->serve_cgi = serve_cgi;
 
+    conn->closed = conn_closed;
     conn->close = conn_close;
 
     conn->incref = conn_incref;
