@@ -40,6 +40,8 @@
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
+#include <string.h>
+
 #include "ssl.h"
 
 #if defined(HAVE_WOLFSSL)
@@ -128,11 +130,27 @@
 struct ssl_context {
 };
 
-int ssl_err_code;
+static int ssl_err_code;
 
-char *ssl_strerror(int error, char *buffer, int len)
+const char *ssl_last_error_string(char *buf, int len)
 {
-    return ERR_error_string(error, buffer);
+    const char *file, *data;
+    int line, flags;
+
+    if (ssl_err_code == SSL_ERROR_SSL) {
+        int used;
+
+        ssl_err_code = ERR_peek_error_line_data(&file, &line, &data, &flags);
+        ERR_error_string_n(ssl_err_code, buf, len);
+
+        used = strlen(buf);
+
+        snprintf(buf + used, len - used, ":%s:%d:%s", file, line, (flags & ERR_TXT_STRING) ? data : "");
+    } else {
+        ERR_error_string_n(ssl_err_code, buf, len);
+    }
+
+    return buf;
 }
 
 struct ssl_context *ssl_context_new(bool server)
@@ -261,8 +279,12 @@ void *ssl_session_new(struct ssl_context *ctx, int sock)
 {
     void *ssl = SSL_new((void *)ctx);
 
-    if (ssl)
-        SSL_set_fd(ssl, sock);
+    if (!ssl)
+        return NULL;
+
+    SSL_set_fd(ssl, sock);
+
+    SSL_set_mode(ssl, SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER | SSL_MODE_ENABLE_PARTIAL_WRITE);
 
     return ssl;
 }
@@ -342,8 +364,8 @@ static bool handle_wolfssl_asn_error(void *ssl, int r,
 }
 #endif
 
-int ssl_connect(void *ssl, bool server,
-        void (*on_verify_error)(int error, const char *str, void *arg), void *arg)
+static int ssl_handshake(void *ssl, bool server,
+    void (*on_verify_error)(int error, const char *str, void *arg), void *arg)
 {
     int r;
 
@@ -375,6 +397,16 @@ int ssl_connect(void *ssl, bool server,
     return SSL_ERROR;
 }
 
+int ssl_accept(void *ssl, void (*on_verify_error)(int error, const char *str, void *arg), void *arg)
+{
+    return ssl_handshake(ssl, true, on_verify_error, arg);
+}
+
+int ssl_connect(void *ssl, void (*on_verify_error)(int error, const char *str, void *arg), void *arg)
+{
+    return ssl_handshake(ssl, false, on_verify_error, arg);
+}
+
 int ssl_write(void *ssl, const void *buf, int len)
 {
     int ret;
